implementation module ppm
import StdEnv, utils, rc, stats, see

:: CxTree a = Leaf a
            | Nothing
            | LargeNode a .{.CxTree a} 
            | CxDet a !Char .(CxTree a)

:: ContextTree :== CxTree .Stats

baseTree = Leaf (StLarge ones 0 256) where
  ones = createArray 256 1

dummyArray :: {#Int}
dummyArray =: createArray 256 1 

createBranch :: {#Char} !Int -> *(CxTree *Stats) 
createBranch context i 
  | i==0 = Leaf (createStats context.[i+1])
  # br = createBranch context (i-1)
  # chld_arr = {Nothing \\ i <-: dummyArray}
  = CxDet (createStats context.[i+1]) context.[i-1] br

estimateSymbol :: !Char !{#Char} !Int !*{#Int} !*ContextTree -> *(Bool,[(!Int,!Int,!Int,!Int,!Bool)],.{#Int},*(CxTree *Stats))
estimateSymbol c cx ci mask Nothing = abort "Nothing in estimateSymbol"

estimateSymbol c cx ci mask (Leaf st) 
  # (found, cumFreq, freq, totFreq, ne, mask, st) = getSymbInfo c mask st
  # st = incrCount c found st
  | ci < 0 = (found, [(cumFreq, freq, totFreq,ne,found)], mask, Leaf st)
  # br = createBranch cx ci //create branch: here ci>=0
  # tr = CxDet st cx.[ci] br
  = (found, [(cumFreq, freq, totFreq,ne,found)], mask, tr)
  
estimateSymbol c cx ci mask (CxDet st cc chld)
  | ci < 0  
    # (found, cumFreq, freq, totFreq,ne, mask, st) = getSymbInfo c mask st
    # st = incrCount c found st
    = (found, [(cumFreq, freq, totFreq,ne,found)], mask, CxDet st cc chld)
  | cx.[ci]==cc //have a higher order context  
    # (found, est_list, mask, chld_tree) = estimateSymbol c cx (ci-1) mask chld
    | found 
       //# st = incrCount c found st //update exclusion
       # tr = CxDet st cc chld_tree
       = (found, est_list, mask, tr)
    //symbol not found in higher order context (child), est_list contains Esc estimations
    # (found, cumFreq, freq, totFreq, ne, mask, st) = getSymbInfo c mask st
    # st = incrCount c found st
    # tr = CxDet st cc chld_tree
    = (found, [(cumFreq, freq, totFreq,ne,found) : est_list], mask, tr) 
  // no higher context, need to build a branch  
  # br = createBranch cx ci 
  # chld_arr = {Nothing \\ i <-: dummyArray}
  # chld_arr & [toInt cx.[ci]] = br
  # chld_arr & [toInt cc] = chld
  # (found, cumFreq, freq, totFreq, ne, mask, st) = getSymbInfo c mask st
  # st = incrCount c found st   
  # tr = LargeNode st chld_arr    
  = (found, [(cumFreq, freq, totFreq,ne,found)], mask, tr)   

estimateSymbol c cx ci mask (LargeNode st chld_arr) 
  | ci < 0
    # (found, cumFreq, freq, totFreq, ne, mask, st) = getSymbInfo c mask st
    # st = incrCount c found st
    = (found, [(cumFreq, freq, totFreq, ne,found)], mask, LargeNode st chld_arr)
  #! chi = toInt cx.[ci]
  #! (chld, chld_arr) = replace chld_arr chi Nothing 
  = case chld of
      Nothing 
        # br = createBranch cx ci 
        # chld_arr & [chi] = br
        # (found, cumFreq, freq, totFreq, ne, mask, st) = getSymbInfo c mask st
        # st = incrCount c found st   
        # tr = LargeNode st chld_arr    
        = (found, [(cumFreq, freq, totFreq, ne,found)], mask, tr) 
      chld_tree 
        # (found, est_list, mask, chld_tree) = estimateSymbol c cx (ci-1) mask chld_tree
        # chld_arr & [chi] = chld_tree        
        | found 
          //# st = incrCount c found st  //update exclusion
          # tr = LargeNode st chld_arr     
          = (found, est_list, mask, tr)
        //symbol not found in higher order context (child), est_list contains Esc estimations
        # (found, cumFreq, freq, totFreq, ne, mask, st) = getSymbInfo c mask st
        # st = incrCount c found st  
        # tr = LargeNode st chld_arr     
        = (found, [(cumFreq, freq, totFreq, ne,found) : est_list], mask, tr) 

updateTree :: !Char {#Char} !Int !Int !*ContextTree -> *ContextTree
updateTree c cx ci depth Nothing = abort "Nothing in updateTree"
updateTree c cx ci depth (Leaf st)
  # found = depth >= 0
  # st = incrCount c found st
  | ci < 0 = Leaf st
  # br = createBranch cx ci 
  = CxDet st cx.[ci] br
  
updateTree c cx ci depth (CxDet st cc chld)
  # found = depth >= 0
  # st = if (depth <= 0) (incrCount c found st) st 
  | ci < 0 = CxDet st cc chld
  | cx.[ci] == cc 
    # chld_tree = updateTree c cx (ci-1) (depth-1) chld
    = CxDet st cc chld_tree
  # br = createBranch cx ci
  # chld_arr = {Nothing \\ i <-: dummyArray}
  # chld_arr & [toInt cx.[ci]] = br
  # chld_arr & [toInt cc] = chld
  = LargeNode st chld_arr   
  
updateTree c cx ci depth (LargeNode st chld_arr)
  # found = depth >= 0
  # st = if (depth <= 0) (incrCount c found st) st 
  | ci < 0 = LargeNode st chld_arr
  #! chi = toInt cx.[ci]
  #! (chld, chld_arr) = replace chld_arr chi Nothing 
  # ch_tr = case chld of
              Nothing = createBranch cx ci
              ch_tree = updateTree c cx (ci-1) (depth-1) ch_tree
  # chld_arr & [chi] = ch_tr
  = LargeNode st chld_arr    
    
decodeSymbolWithStats ::  *RCState *{#.Int} *Stats EscCtx *EscTab -> *(*RCState,!Int,.{#Int},*Stats,.EscTab)
decodeSymbolWithStats rc mask st ecx etab
  #! (mask, st, nc, ne) = getTotFreq mask st
  #! (neSd, ncSd, shift, etab) = getEscN ne nc ecx etab 
  #! totFr = neSd + ncSd
  #! (valSd, rc) = getFreq rc totFr 
  #! val = if (valSd < neSd) 0 (rot (valSd - neSd) (~shift) + ne)
  #! (symb, cumFr, fr, mask, st) = findSymbByVal val mask st 
  #! esc = symb < 0 
  #! freq = if esc neSd (rot fr shift)
  #! cumFreq = if esc 0 (rot (cumFr - ne) shift + neSd)  
  #! etab = updEscN ne nc ecx etab esc 
  #! rc = decode rc cumFreq freq
  = (rc, symb, mask, st, etab)

//does not update stats or build new branches, must be done later
// rc cx ci depth mask ecx etab tr -> (rc, symb, found_depth, nesc, mask,etab, tr)
decodeSymbol :: *RCState {#Char} !Int !Int *{#.Int} EscCtx *EscTab *ContextTree -> *(*RCState,!Int,!Int,!Int,*{#Int},.EscTab,*ContextTree)
decodeSymbol rc cx ci depth mask ecx etab Nothing = abort "Nothing in decodeSymbol"
decodeSymbol rc cx ci depth mask ecx etab tr=:(Leaf st) 
  # (rc, symb, mask, st, etab) = decodeSymbolWithStats rc mask st ecx etab
  = (rc, symb, depth, (if (symb >= 0) 0 1), mask, etab, Leaf st)
  
decodeSymbol rc cx ci depth mask ecx etab tr=:(CxDet st cc chld)  
  | ci < 0 
    # (rc, symb, mask, st, etab) = decodeSymbolWithStats rc mask st ecx etab
    = (rc, symb, depth, (if (symb >= 0) 0 1), mask, etab, CxDet st cc chld)
  | cx.[ci]==cc
      # (rc, symb, dpt, escn, mask, etab, ch_tr) = decodeSymbol rc cx (ci-1) (depth+1) mask ecx etab chld
      | symb >=0 = (rc, symb, dpt,escn, mask, etab, CxDet st cc ch_tr)
      # (rc, symb, mask, st, etab) = decodeSymbolWithStats rc mask st ecx etab
      = (rc, symb, depth,escn + (if (symb >= 0) 0 1), mask, etab, CxDet st cc ch_tr)
  # (rc, symb, mask, st, etab) = decodeSymbolWithStats rc mask st ecx etab
  = (rc, symb, depth,(if (symb >= 0) 0 1), mask, etab, CxDet st cc chld)    
  
decodeSymbol rc cx ci depth mask ecx etab tr=:(LargeNode st chld_arr)
  | ci < 0 
    # (rc, symb, mask, st, etab) = decodeSymbolWithStats rc mask st ecx etab
    = (rc, symb, depth, (if (symb >= 0) 0 1), mask, etab, LargeNode st chld_arr)
  #! chi = toInt cx.[ci]
  #! (chld, chld_arr) = replace chld_arr chi Nothing 
  = case chld of
      Nothing 
        # (rc, symb, mask, st, etab) = decodeSymbolWithStats rc mask st ecx etab       
        = (rc, symb, depth, (if (symb >= 0) 0 1), mask, etab, LargeNode st chld_arr)
      chld_tree
        # (rc, symb, dpt,escn, mask, etab, ch_tr) = decodeSymbol rc cx (ci-1) (depth+1) mask ecx etab chld_tree
        # chld_arr & [chi] = ch_tr //put back wat was replace'd just before the case
        | symb >=0 = (rc, symb, dpt,escn, mask, etab, LargeNode st chld_arr)
        # (rc, symb, mask, st, etab) = decodeSymbolWithStats rc mask st ecx etab
        = (rc, symb, depth, escn + (if (symb >= 0) 0 1), mask, etab, LargeNode st chld_arr)     

Order :== 4

ppm_compress :: {#Char} -> .{#Char}
ppm_compress text = finishEncode rc where
  rc0 = newEncoder (20 + size text)
  tr0 :: *ContextTree
  tr0 = baseTree
  (rc,tr,_,_,_,etab) = foldArray encodeSymb (rc0,tr0,"",mask0, 1, etab0) text  
  mask0 = createArray 256 1
  etab0 = mk_etab
  encodeSymb (rc,tr,cx,mask, prev_succ, etab) c
    # cxlen = size cx
    # cx = (if (cxlen < Order+1) cx (cx % (1, cxlen)) ) +++ {c}
    #! mask = { mask & [i] = 1 \\ i <- [0..255]}
    # (found, est_list, mask, tr) = estimateSymbol c cx (size cx - 2) mask tr
    # success = if (length est_list == 1) 1 0
    # (rc,etab,_) = foldr encodeInterv (rc,etab,prev_succ) est_list      
    = (rc, tr, cx, mask, success, etab)
        
  encodeInterv ival=:(cf,fr,tf, ne,found) (rc,etab,ecx)
    #! ((cumFreq, freq, totFreq),etab) = reestimateEsc ecx etab ival 
    #! rc = encode rc cumFreq freq totFreq
    #! etab = updEscN ne (tf-ne) ecx etab (not found)
    = (rc,etab,ecx)
      
ppm_decompress :: {#Char} !Int -> .{#Char}      
ppm_decompress compressed len = decloop 0 rc0 tr0 buf0 mask0 "" 1 etab0 where
  rc0 = newDecoder compressed
  buf0 :: *{#Char}
  buf0 = createArray len '-'
  tr0 = baseTree  
  etab0 = mk_etab
  mask0 = createArray 256 1
  decloop pos rc tr buf mask cx ecx etab  
    | pos >= len = buf
    #! mask = { mask & [i] = 1 \\ i <- [0..255]}
    # (rc, symb, depth, escn, mask, etab, tr) = decodeSymbol rc cx (size cx - 1) 0 mask ecx etab tr    
    # success = if (escn == 0) 1 0
    # c = toChar symb
    # buf & [pos] = c
    # cx = cx +++ {c}
    # cxlen = size cx    
    # tr = updateTree c cx (cxlen - 2) depth tr
    # cx = (if (cxlen <= Order) cx (cx % (1, cxlen)) ) 
    = decloop (pos+1) rc tr buf mask cx success etab 
