implementation module rc
import StdEnv, utils

:: *Buffer :== ( *{#Char} , !Int)

writeBytes [] buf = buf
writeBytes [b:bs] (buffer, pos) 
  #! buf = ({buffer & [pos] = b}, pos+1)
  = writeBytes bs buf

bufferContents :: *Buffer -> .{#Char}
bufferContents (buf,pos) 
  | pos > 0 = { buf.[i] \\ i <- [0..pos-1] }
            = ""

newBuffer :: Int -> *Buffer
newBuffer sz = (createArray sz '-', 0)

readByte :: *Buffer -> (!Char, *Buffer)
readByte (buf, pos) = (ch, (buf1, pos+1)) where
  (ch, buf1) = buf![pos]

:: *RCState = { buf :: !*Buffer
              , low :: !Int
              , range :: !Int
              , val :: !Int
              , ffnum :: !Int
              , cache :: !Int
              , carry :: !Int
              }
      
TOP :== 0x10000
Thres :== 255*TOP
            
newEncoder :: Int -> RCState
newEncoder sz = { buf = newBuffer sz, low = 0, 
  range = 0xFFFFFF, val = 0, ffnum = 0, cache = 0, carry = 0 }
    
encode :: !*RCState !.Int !.Int !.Int -> *RCState    
encode rc cumFreq freq totFreq 
  = renorm rc1 where
    quant = rc.range / totFreq
    low1 = (rc.low + cumFreq * quant) bitand 0xFFFFFF
    rc1  = { rc & low = low1
         , range = quant * freq
         , carry = if (low1 < rc.low) (rc.carry+1) rc.carry
         }
  
renorm rc 
  | rc.range < TOP = renorm (shiftLow rc)
                   = rc

shiftLow rc = if (rc.low < Thres || rc.carry > 0) rc1 rc2 where
  range1 = rc.range << 8  
  low1 = (rc.low bitand 0xFFFF) << 8
  rc1 = { rc & range = range1, low = low1, ffnum = 0, cache = rc.low >> 16,
          buf = buf1, carry = 0 }  
  buf1 = writeBytes [toChar (rc.cache + rc.carry) : repeatn rc.ffnum (toChar (rc.carry-1))] rc.buf
  rc2 = { rc & low = low1, range = range1, ffnum = rc.ffnum+1 }  
  
finishEncode :: *RCState -> .{#Char}  
finishEncode rc = bufferContents rc1.buf where
  rc1 = iter 4 shiftLow rc  

newDecoder :: {#Char} -> *RCState
newDecoder str = iter 4 loadByte rc where
  rc = { buf = (buf,0), low = 0, range = 0xFFFFFF, 
         val = 0, ffnum = 0, cache = 0, carry = 0 }
  buf = {c \\ c <-: str}
  
loadByte rc 
  # (c, buf1) = readByte rc.buf
  = { rc & buf = buf1, val = ((rc.val bitand 0xFFFF) << 8) + toInt c }

getFreq :: !*RCState !.Int -> *(Int,*RCState)
getFreq rc totFreq 
  #! rc = { rc & range = rc.range / totFreq }
  #! freq = rc.val / rc.range
  = (freq, rc)

decode :: !*RCState !.Int !.Int -> *RCState
decode rc cumFreq freq = load rc1 where
  rc1 = { rc & val = rc.val - cumFreq * rc.range, range = rc.range * freq }
  load rc
    | rc.range < TOP = load (loadByte {rc & range = rc.range << 8})
                     = rc
