implementation module stats
import StdEnv, utils

SmallSize :== 10
MaxTotal :== 65535
NE :== 6
DC :== 11
DE :== 5

:: Stats = StDet !Char !Int // c, Nc. Nesc=NE
         | StSmall !String .{#Int} !Int !Int //symbols, counters, Nesc, total
         | StLarge .{#Int} !Int !Int // counters, Nesc, total

createStats :: !Char  -> .Stats
createStats c = StDet c 1
    
incrCount :: !Char !Bool !*Stats -> *Stats 
incrCount ch True st = norm (updateStats ch DC 0 st) 
incrCount ch False st = norm (updateStats ch DC DE st) 

norm :: !*Stats -> *Stats         
norm st=:(StDet c nc)
  | nc+NE <= MaxTotal = st
  = StDet c (nc/2+1) 
  
norm st=:(StSmall symbs tab ne tot)
  | tot <= MaxTotal = st
  #! tab1 = {x/2+1 \\ x <-: tab}
     ne1 = ne/2+1
     tot1 = foldArray (+) ne1 tab1
  = StSmall symbs tab1 ne1 tot1
       
norm st=:(StLarge tab ne tot)
  | tot <= MaxTotal = st
  #! tab1 = { if (x>0) (x/2+1) 0 \\ x <-: tab }
     ne1 = ne/2+1
     tot1 = foldArray (+) ne1 tab1
  = StLarge tab1 ne1 tot1

getTotFreq :: !u:{#Int} !w:Stats -> (x:{#Int},z:Stats,!Int,!Int), [u <= x, w <= z]
getTotFreq mask st=:(StDet c nc) 
  #! t = nc * mask.[toInt c] 
  = (mask, st, t, NE)
  
getTotFreq mask st=:(StSmall symbs tab ne tot) 
  #! t = gtfloop2 0 0 // sum [tab.[i] * mask.[toInt symbs.[i]] \\ i <- [0..size symbs - 1]] 
  = (mask, st, t, ne) where
    n = size symbs - 1
    gtfloop2 i sum 
      | i > n = sum
      = gtfloop2 (i+1) (sum + tab.[i] * mask.[toInt symbs.[i]])

getTotFreq mask st=:(StLarge tab ne tot) 
  #! t = gtfloop 0 0 //ne + sum [mask.[i] * tab.[i] \\ i <- [0..255]]
  = (mask, st, t, ne) where
    gtfloop i sum 
      | i > 255 = sum
      = gtfloop (i+1) (sum + tab.[i] * mask.[i])
         
calcCumFr :: !.Int {#.Int} {#.Int} -> Int         
calcCumFr n mask tab = cfloop 0 0 where
  cfloop i cf
    | i == n = cf
    #! cf1 = cf + tab.[i] * mask.[i]
    = cfloop (i+1) cf1 

calcCumFrStr :: !.Int {#.Int} {#.Int} .{#Char} -> Int   
calcCumFrStr n mask tab str = cfsloop 0 0 where
  cfsloop i cf
    | i == n = cf
    #! cf1 = cf + tab.[i] * mask.[toInt str.[i]]
    = cfsloop (i+1) cf1    
    
updateMask :: *{#Int} u:{#Int} -> *(*{#Int},v:{#Int}), [u <= v]
updateMask mask tab = umloop 0 mask tab where
  umloop i m tab
    | i > 255 = (m, tab)
    | tab.[i] >0 =  umloop (i+1) {m & [i] = 0} tab 
    = umloop (i+1) m tab     
    
updateMaskStr :: *{#Int} String -> *{#Int}
updateMaskStr mask symbs = umsloop 0 mask where
  umsloop i m
    | i == n = m
    = umsloop (i+1) {m & [toInt symbs.[i]] = 0}
  n = size symbs
   
//estimate symbol and update mask if symbol not found
getSymbInfo :: !.Char *{#Int} !u:Stats -> (.Bool,!Int,!Int,!Int,!Int,*{#Int},v:Stats), [u <= v]
getSymbInfo ch mask st=:(StDet c nc) 
  | c==ch = (True, NE, nc, NE+nc, NE, mask, st)
  # ci = toInt c
  # (mci, mask) = mask![ci]
  # mask & [ci] = 0
  = (False, 0, NE, NE+nc*mci, NE, mask, st) 

getSymbInfo ch mask st=:(StLarge tab ne tot)
  # i = toInt ch
  | tab.[i] > 0 
     #! cf = sumloop 0 0 (i-1)//sum [tab.[j] * mask.[j] \\ j <- [0..i-1]] //calcCumFr i mask tab
     #! rest = sumloop 0 i 255//sum [tab.[j] * mask.[j] \\ j <- [i..255]]
     #! fr = tab.[i]
     #! totFr = ne+cf+rest
     = (True, cf+ne, fr, totFr, ne, mask, st)
  #! totFr = sumloop ne 0 255//ne + sum [tab.[j] * mask.[j] \\ j <- [0..255]]   
  #! (mask, tab) = updateMask mask tab
  #! st = StLarge tab ne tot
  = (False, 0, ne, totFr, ne, mask, st) where
    sumloop sum i last
      | i > last = sum
      = sumloop (sum + tab.[i] * mask.[i]) (i+1) last
  
getSymbInfo ch mask st=:(StSmall symbs tab ne tot)
  #! j = findChar ch symbs
  #! totFr = sumloop2 ne 0//ne + sum[tab.[j] * mask.[toInt symbs.[j]] \\ j <- [0..n-1]] //!
  | j < 0     
      #! mask = updateMaskStr mask symbs      
      = (False, 0, ne, totFr, ne, mask, st)
  #! cf = calcCumFrStr j mask tab symbs
  #! fr = tab.[j] 
  = (True, cf+ne, fr, totFr, ne, mask, st) where
    n = size symbs
    sumloop2 sum i     
      | i >= n = sum
      = sumloop2 (sum + tab.[i] * mask.[toInt symbs.[i]]) (i+1) 
  
//inverse to estimation, update mask if this is Esc
//val mask st -> (c or -1, cumFreq, freq, mask, st)
findSymbByVal :: Int *{#Int} u:Stats -> *(Int,Int,Int,*{#Int},v:Stats), [u <= v]
findSymbByVal val mask st=:(StDet c nc) 
  | val < NE 
    # mask & [toInt c] = 0
    = (-1, 0, NE, mask, st)  
  = (toInt c, NE, nc, mask, st)

findSymbByVal val mask st=:(StLarge tab ne tot) 
  | val < ne 
    # (mask, tab) = updateMask mask tab
    = (-1, 0, ne, mask, StLarge tab ne tot)
  = fsloop 0 ne tab  where
    fsloop i cumfr tab
      #! nf = cumfr + tab.[i] * mask.[i]
      #! fr = tab.[i]
      | nf > val  = (i, cumfr, fr, mask, StLarge tab ne tot)
      = fsloop (i+1) nf tab
 
findSymbByVal val mask st=:(StSmall symbs tab ne tot) 
  | val < ne 
    # mask = updateMaskStr mask symbs
    = (-1, 0, ne, mask, st)
  = fsloop2 0 ne tab  where
    fsloop2 i cumfr tab 
      #! nf = cumfr + tab.[i] * mask.[toInt symbs.[i]]
      #! fr = tab.[i]
      | nf > val = (toInt symbs.[i], cumfr, fr, mask, StSmall symbs tab ne tot)
      = fsloop2 (i+1) nf tab 
    
updateStats :: !Char !Int !Int !*Stats -> *Stats 
updateStats ch deltaC deltaEsc (StLarge tab ne tot)
  #! i = toInt ch
  #! old = tab.[i]
  = StLarge {tab & [i] = old + deltaC} (ne+deltaEsc) (tot+deltaC+deltaEsc)
  
updateStats ch deltaC deltaEsc (StDet c nc) 
  | c==ch = StDet c (nc + deltaC) 
  | c < ch = mksmall c ch nc deltaC
  = mksmall ch c deltaC nc
 where
  mksmall c1 c2 n1 n2 = StSmall {c1, c2} {n1, n2} (NE+deltaEsc) (n1+n2+NE+deltaEsc)

updateStats ch deltaC deltaEsc (StSmall symbs tab ne tot)
  #! j = findChar ch symbs
  | j >= 0 
      #! old = tab.[j]
      = StSmall symbs {tab & [j] = old + deltaC} (ne+deltaEsc) (tot + deltaC + deltaEsc)
  | size symbs >= SmallSize
      = StLarge ltab (ne+deltaEsc) (tot + deltaC+deltaEsc) with
          ltab = foldArray f {a0 & [toInt ch] = deltaC} pairs
          a0 :: *{#Int}
          a0 = createArray 256 0
          pairs :: {(Char,Int)}
          pairs = {(c,cnt) \\ c <-: symbs & cnt <-: tab}
          f a (c,cnt) = {a & [toInt c] = cnt}
  = StSmall symbs1 tab1 (ne+deltaEsc) (tot+deltaC+deltaEsc) with
      (symbs1, tab1) = insert_loop 0 symbs0 tab0
      symbs0 :: *{#Char}
      symbs0 = createArray (n+1) '-'
      tab0 :: *{#Int}
      tab0 = createArray (n+1) 0
      n = size symbs
      insert_loop i s t
        | i >= n = ({s & [i] = ch}, {t & [i] = deltaC})
        | symbs.[i] < ch = insert_loop (i+1) {s & [i] = symbs.[i]} {t & [i] = tab.[i]}
        = insert_end_loop i {s & [i] = ch} {t & [i] = deltaC}
      insert_end_loop i s t
        | i >= n = (s, t)
        = insert_end_loop (i+1) {s & [i+1] = symbs.[i]} {t & [i+1] = tab.[i]}
  